/*
   Copyright (C) 2002-2019 CERN for the benefit of the ATLAS collaboration
   */

/**   @file TrackerEfficiencyAlg.cxx
 *   Implementation file for the TrackerEfficiencyAlg class.
 *   @author Savannah Shively
 *   @date 2 September 2021
 */
#include "TrackerEfficiencyAlg.h"
#include "Identifier/Identifier.h"
#include "StoreGate/ReadHandle.h"
#include "StoreGate/StoreGateSvc.h"
#include "TrackerPrepRawData/TrackerClusterCollection.h"
#include "TrackerPrepRawData/FaserSCT_Cluster.h"
#include "TrackerPrepRawData/FaserSCT_ClusterCollection.h"  //new!
#include "TrackerIdentifier/FaserSCT_ID.h"
#include "TrkTrack/Track.h"
#include "TrkTrack/TrackStateOnSurface.h"

#include "TrackerSpacePoint/FaserSCT_SpacePointContainer.h"
#include "TrackerRIO_OnTrack/FaserSCT_ClusterOnTrack.h"


#include <TTree.h>
#include <TBranch.h>
#include <numeric>
#include <math.h>  


// Constructor with parameters:
TrackerEfficiencyAlg::TrackerEfficiencyAlg(const std::string &name, ISvcLocator* pSvcLocator) :
  AthReentrantAlgorithm(name, pSvcLocator),  AthHistogramming(name), m_idHelper(nullptr), m_histSvc("THistSvc/THistSvc", name) {}

// Initialize method:
StatusCode TrackerEfficiencyAlg::initialize() {
  //ATH_CHECK(histSvc()->regHist("THistSvc/THistSvc", m_histSvc));
  //ATH_MSG_ALWAYS("Initializing...");

  ATH_CHECK( m_trackCollection.initialize() );
  ATH_CHECK( m_clusterContainerKey.initialize() );
  ATH_CHECK( m_waveformPreshowerHitContainerKey.initialize() );
  ATH_CHECK( m_waveformCaloHitContainerKey.initialize() );
  ATH_CHECK( m_waveformVetoHitContainerKey.initialize() );
  ATH_CHECK(detStore()->retrieve(m_idHelper, "FaserSCT_ID"));
  ATH_CHECK(m_spacePointContainerKey.initialize());

  //Deion code
  ATH_CHECK(m_clusterContainer.initialize());
   ATH_CHECK(detStore()->retrieve(m_sctHelper,       "FaserSCT_ID"));

  m_tree = new TTree("trackCounts", "Track Counts");

  //Raw Integral
  m_tree->Branch("presh_rawintegral", &m_presh_rawintegral, "presh_rawintegral/F");
  m_tree->Branch("calo_rawintegral", &m_calo_rawintegral, "calo_rawintegral/F");
  m_tree->Branch("veto_rawintegral", &m_veto_rawintegral, "veto_rawintegral/F");

  m_tree->Branch("presh_integral", &m_presh_integral, "presh_integral/F");
  m_tree->Branch("calo_integral", &m_calo_integral, "calo_integral/F");
  m_tree->Branch("calo_integral_norm", &m_calo_integral_norm, "calo_integral_norm/F");
  m_tree->Branch("veto_integral", &m_veto_integral, "veto_integral/F");

  //Preshower Counts
  m_tree->Branch("presh_wfhit_count", &m_presh_wfhit_count, "presh_wfhit_count/I");
  m_tree->Branch("presh_counts_ch6", &m_presh_counts_ch6, "presh_counts_ch6/I");
  m_tree->Branch("presh_counts_ch7", &m_presh_counts_ch7, "presh_counts_ch7/I");
  
  //Calorimeter Counts
  m_tree->Branch("calo_wfhit_count", &m_calo_wfhit_count, "calo_wfhit_count/I");
  m_tree->Branch("calo_counts_ch0", &m_calo_counts_ch0, "calo_counts_ch0/I");
  m_tree->Branch("calo_counts_ch1", &m_calo_counts_ch1, "calo_counts_ch1/I");
  m_tree->Branch("calo_counts_ch2", &m_calo_counts_ch2, "calo_counts_ch2/I");
  m_tree->Branch("calo_counts_ch3", &m_calo_counts_ch3, "calo_counts_ch3/I");
  m_tree->Branch("calo_counts_ch4", &m_calo_counts_ch4, "calo_counts_ch4/I");
  m_tree->Branch("calo_counts_ch5", &m_calo_counts_ch5, "calo_counts_ch5/I");

  //Veto Counts
  m_tree->Branch("veto_wfhit_count", &m_veto_wfhit_count, "veto_wfhit_count/I");
  m_tree->Branch("veto_counts_ch8", &m_veto_counts_ch8, "veto_counts_ch8/I");
  m_tree->Branch("veto_counts_ch9", &m_veto_counts_ch9, "veto_counts_ch9/I");
  
  //More general counts
  m_tree->Branch("track_counts", &m_track_counts, "track_counts/I");
  m_tree->Branch("event_counts", &m_event_counts, "event_counts/I");
  m_tree->Branch("cluster_counts", &m_cluster_counts, "cluster_counts/I");
  m_tree->Branch("trcluster counts", &m_trcluster_counts, "trcluster_counts/I");

  //Counting how many channels are active 
  m_tree->Branch("presh_max_channel", &m_presh_max_channel, "presh_max_channel/I");
  m_tree->Branch("calo_max_channel", &m_calo_max_channel, "calo_max_channel/I");
  m_tree->Branch("veto_max_channel", &m_veto_max_channel, "veto_max_channel/I");

  m_tree->Branch("calo_sumintegral", &m_calo_sumintegral, "calo_sumintegral/I");

  //Checking that all channels are active
  m_tree->Branch("presh_check", &m_presh_check, "presh_check/I");
  m_tree->Branch("calo_check", &m_calo_check, "calo_check/I");
  m_tree->Branch("veto_check", &m_veto_check, "veto_check/I");

  m_tree->Branch("presh_alpha", &m_presh_alpha, "presh_alpha/F");
  m_tree->Branch("calo_alpha", &m_calo_alpha, "calo_alpha/F");
  m_tree->Branch("veto_alpha", &m_veto_alpha, "veto_alpha/F");

  //Tracker data
  m_tree->Branch("x", &m_x, "x/F");
  m_tree->Branch("y", &m_y, "y/F");
  m_tree->Branch("z", &m_z, "z/F");
  m_tree->Branch("px", &m_px, "px/F");
  m_tree->Branch("py", &m_py, "py/F");
  m_tree->Branch("pz", &m_pz, "pz/F");
  m_tree->Branch("eta", &m_eta, "eta/F");
  m_tree->Branch("theta", &m_theta, "theta/F");
  m_tree->Branch("pt", &m_pt, "pt/F");


  m_tree->Branch("chi_squared", &m_chi2, "chi2/D");
  ATH_CHECK(histSvc()->regTree("/HIST/trackCounts", m_tree));

  //the rest of the waveform data
  m_tree->Branch("presh_width", &m_presh_width, "presh_width/F");
  m_tree->Branch("presh_localtime", &m_presh_localtime, "presh_localtime/F");
  m_tree->Branch("presh_peak", &m_presh_peak, "presh_peak/F");
  m_tree->Branch("presh_bcid_time", &m_presh_bcid_time, "presh_bcid_time/F");
  m_tree->Branch("presh_raw_peak", &m_presh_raw_peak, "presh_raw_peak/F");
  m_tree->Branch("presh_baseline_mean", &m_presh_baseline_mean, "presh_baseline_mean/F");
  m_tree->Branch("presh_baseline_rms", &m_presh_baseline_rms, "presh_baseline_rms/F");

  m_tree->Branch("calo_width", &m_calo_width, "calo_width/F");
  m_tree->Branch("calo_localtime", &m_calo_localtime, "calo_localtime/F");
  m_tree->Branch("calo_peak", &m_calo_peak, "calo_peak/F");
  m_tree->Branch("calo_bcid_time", &m_calo_bcid_time, "calo_bcid_time/F");
  m_tree->Branch("calo_raw_peak", &m_calo_raw_peak, "calo_raw_peak/F");
  m_tree->Branch("calo_baseline_mean", &m_calo_baseline_mean, "calo_baseline_mean/F");
  m_tree->Branch("calo_baseline_rms", &m_calo_baseline_rms, "calo_baseline_rms/F");

  m_tree->Branch("calo_integral2", &m_calo_integral2, "calo_integral2/F");
  m_tree->Branch("calo_max_channel2", &m_calo_max_channel2, "calo_max_channel2/I");
  m_tree->Branch("calo_width2", &m_calo_width2, "calo_width2/F");
  m_tree->Branch("calo_localtime2", &m_calo_localtime2, "calo_localtime2/F");
  m_tree->Branch("calo_peak2", &m_calo_peak2, "calo_peak2/F");
  m_tree->Branch("calo_bcid_time2", &m_calo_bcid_time2, "calo_bcid_time2/F");
  m_tree->Branch("calo_raw_peak2", &m_calo_raw_peak2, "calo_raw_peak2/F");
  m_tree->Branch("calo_baseline_mean2", &m_calo_baseline_mean2, "calo_baseline_mean2/F");
  m_tree->Branch("calo_baseline_rms2", &m_calo_baseline_rms2, "calo_baseline_rms2/F");

  m_tree->Branch("veto_width", &m_veto_width, "veto_width/F");
  m_tree->Branch("veto_localtime", &m_veto_localtime, "veto_localtime/F");
  m_tree->Branch("veto_peak", &m_veto_peak, "veto_peak/F");
  m_tree->Branch("veto_bcid_time", &m_veto_bcid_time, "veto_bcid_time/F");
  m_tree->Branch("veto_raw_peak", &m_veto_raw_peak, "veto_raw_peak/F");
  m_tree->Branch("veto_baseline_mean", &m_veto_baseline_mean, "veto_baseline_mean/F");
  m_tree->Branch("veto_baseline_rms", &m_veto_baseline_rms, "veto_baseline_rms/F");


  //cluster
  m_tree->Branch("LayerClusters", &m_layerclusters);

  m_tree->Branch("SpacePoints",&m_nspacepoints);
  m_tree->Branch("LayerSpacePoints", &m_layerspacepoints);


  return StatusCode::SUCCESS;
}

StatusCode TrackerEfficiencyAlg::execute(const EventContext& ctx) const
{
  m_event_counts=0;
  //ATH_MSG_ALWAYS("Executing...");
  //Check that all the events in the file are readable

  //PREPARE AND CHECK READHANDLES
  SG::ReadHandle<Tracker::FaserSCT_ClusterContainer> SCT_ClusterContainer {m_clusterContainerKey, ctx};
  SG::ReadHandle<xAOD::WaveformHitContainer> waveformPreshowerHitContainer {m_waveformPreshowerHitContainerKey, ctx};
  SG::ReadHandle<xAOD::WaveformHitContainer> waveformCaloHitContainer {m_waveformCaloHitContainerKey, ctx};
  SG::ReadHandle<xAOD::WaveformHitContainer> waveformVetoHitContainer {m_waveformVetoHitContainerKey, ctx};
  //SG::ReadHandleKey<FaserSCT_SpacePointContainer> spacePointContainer {m_Sct_spcontainerKey, ctx};

  if (!waveformPreshowerHitContainer.isValid()){
    ATH_MSG_ERROR("Could not read " << waveformPreshowerHitContainer.name());
    return StatusCode::FAILURE;
    }
  if (!waveformCaloHitContainer.isValid()){
    ATH_MSG_ERROR("Could not read " << waveformCaloHitContainer.name());
    return StatusCode::FAILURE;
    }

  if (!waveformVetoHitContainer.isValid()){
    ATH_MSG_ERROR("Could not read " << waveformVetoHitContainer.name());
    return StatusCode::FAILURE;
    }

    
  //clear vectors before using them
  m_layerclusters[0].clear();
  m_layerclusters[1].clear();
  m_layerclusters[2].clear();
  m_layerclusters[3].clear();
  m_layerclusters[4].clear();



  //ANALYSE PRESHOWER EVENTS
  m_presh_wfhit_count=0; //counts number of hits in a waveform hit container
  m_presh_counts_ch6=0;  //counts number of hits in a specific channel
  m_presh_counts_ch7=0;
  int presh_check[2] = {0,0};
  float presh_maxintegral=0;
  m_presh_max_channel=-1;

  for (const xAOD::WaveformHit* hit : *waveformPreshowerHitContainer)
  {
    if ((hit->channel() == 6) || (hit->channel() == 7)) 
    {
      if (abs(presh_maxintegral) <abs(hit->integral())){
        const Identifier id = hit->identify();
        m_presh_rawintegral = hit->raw_integral();
        m_presh_integral = hit->integral();
        presh_maxintegral=hit->integral();
        m_presh_alpha = hit->alpha();

        m_presh_localtime = hit->localtime();
        m_presh_width = hit->width();
        m_presh_peak = hit->peak();
        m_presh_bcid_time = hit->bcid_time();
        m_presh_raw_peak = hit->raw_peak();
        m_presh_baseline_mean = hit->baseline_mean();
        m_presh_baseline_rms = hit->baseline_rms();
        m_presh_max_channel=hit->channel(); //stores the channel number of the max-integral hit
      }
      m_presh_wfhit_count++;

      if (hit->channel() == 6) 
      {presh_check[0]=1;
        m_presh_counts_ch6++;}
      if (hit->channel() == 7) 
      {presh_check[1]=1;
        m_presh_counts_ch7++;}
    }
    
  }

  m_cluster_counts=0;
  
  m_presh_check=presh_check[0]+presh_check[1];

  //ANALYSE CALORIMETER EVENTS
  double Calo_ch_normilization_Dict[6] = {0.9692127178894433, 0.47893862313101704, 1.0, 0.8665297898776219,0.8879248622575592,0.7799859504197602};

  m_calo_wfhit_count = 0;
  m_calo_counts_ch0=0;
  m_calo_counts_ch1=0;
  m_calo_counts_ch2=0;
  m_calo_counts_ch3=0;
  m_calo_counts_ch4=0;
  m_calo_counts_ch5=0;
  int calo_check[6] = {0,0,0,0,0,0};
  float calo_maxintegral=0;
  float calo_maxintegral2=0;
  m_calo_sumintegral=0;

  //Set everything to a default value
  m_calo_alpha = -1;
  m_calo_localtime = -1;
  m_calo_width = -1;
  m_calo_peak = -1;
  m_calo_bcid_time = -1;
  m_calo_raw_peak = -1;
  m_calo_baseline_mean =-1;
  m_calo_baseline_rms = -1;
  m_calo_max_channel = -1;
  m_calo_alpha2 = -1;
  m_calo_localtime2 = -1;
  m_calo_width2 = -1;
  m_calo_peak2 = -1;
  m_calo_bcid_time2 = -1;
  m_calo_raw_peak2 = -1;
  m_calo_baseline_mean2 =-1;
  m_calo_baseline_rms2 = -1;
  m_calo_max_channel2 = -1;

  for (const xAOD::WaveformHit* hit : *waveformCaloHitContainer)
  {
    const Identifier id = hit->identify();
    if (hit->integral()<100000) { m_calo_sumintegral += hit->integral()/Calo_ch_normilization_Dict[hit->channel()];} //Sometimes the integral fit is bad
    if (abs(calo_maxintegral) < abs(hit->integral()/Calo_ch_normilization_Dict[hit->channel()])){
      m_calo_rawintegral = hit->raw_integral();
      m_calo_integral = hit->integral();
      m_calo_integral_norm = m_calo_integral/Calo_ch_normilization_Dict[hit->channel()];
      calo_maxintegral = hit->integral()/Calo_ch_normilization_Dict[hit->channel()]; 
      //store the channel with the highest integral for having just one for each event
      //if its bigger, update stuff, otherwise continue
      //should be nonzero calo signal; may not need alpha and width cuts
      m_calo_alpha = hit->alpha();
      m_calo_localtime = hit->localtime();
      m_calo_width = hit->width();
      m_calo_peak = hit->peak();
      m_calo_bcid_time = hit->bcid_time();
      m_calo_raw_peak = hit->raw_peak();
      m_calo_baseline_mean = hit->baseline_mean();
      m_calo_baseline_rms = hit->baseline_rms();
      m_calo_max_channel=hit->channel();
    }
    else if (abs(calo_maxintegral2) <abs(hit->integral()/Calo_ch_normilization_Dict[hit->channel()]))
    {
      m_calo_rawintegral2 = hit->raw_integral();
      m_calo_integral2 = hit->integral();
      m_calo_integral_norm2 = m_calo_integral2/Calo_ch_normilization_Dict[hit->channel()];
      calo_maxintegral2 = hit->integral()/Calo_ch_normilization_Dict[hit->channel()]; 
      m_calo_alpha2 = hit->alpha();
      m_calo_localtime2 = hit->localtime();
      m_calo_width2 = hit->width();
      m_calo_peak2 = hit->peak();
      m_calo_bcid_time2 = hit->bcid_time();
      m_calo_raw_peak2 = hit->raw_peak();
      m_calo_baseline_mean2 = hit->baseline_mean();
      m_calo_baseline_rms2 = hit->baseline_rms();
      m_calo_max_channel2 =hit->channel();
    }

    int hit_chan=hit->channel();
    
    
    m_calo_wfhit_count++;
    calo_check[hit_chan]=1;
    switch(hit_chan){
      case 0:m_calo_counts_ch0++; break;
      case 1:m_calo_counts_ch1++; break;
      case 2:m_calo_counts_ch2++; break;
      case 3:m_calo_counts_ch3++; break;
      case 4:m_calo_counts_ch4++; break;
      case 5:m_calo_counts_ch5++;
    }
  }
  m_calo_check=calo_check[0]+calo_check[1]+calo_check[2]+calo_check[3]+calo_check[4]+calo_check[5];

  //ANALYSE Veto EVENTS
  
  m_veto_wfhit_count=0;
  m_veto_counts_ch8=0;
  m_veto_counts_ch9=0;
  int veto_check[2] = {0,0};
  float veto_maxintegral=0;
  m_veto_max_channel=-1;

  for (const xAOD::WaveformHit* hit : *waveformVetoHitContainer)
  {
    if (abs(veto_maxintegral) <abs(hit->integral())){
      const Identifier id = hit->identify();
      m_veto_rawintegral = hit->raw_integral();
      m_veto_integral = hit->integral();
      veto_maxintegral = hit->integral();
      m_veto_alpha = hit->alpha();

      m_veto_localtime = hit->localtime();
      m_veto_width = hit->width();
      m_veto_peak = hit->peak();
      m_veto_bcid_time = hit->bcid_time();
      m_veto_raw_peak = hit->raw_peak();
      m_veto_baseline_mean = hit->baseline_mean();
      m_veto_baseline_rms = hit->baseline_rms();
      m_veto_max_channel=hit->channel();
    }

    
    m_veto_wfhit_count++;

    switch(hit->channel()){
      case 8:  m_veto_counts_ch8++; veto_check[0]=1; break;
      case 9:  m_veto_counts_ch9++; veto_check[1]=1; break;
    }
  }
  m_veto_check=veto_check[0]+veto_check[1];

  Amg::Vector3D pos;
  Amg::Vector3D dir;
  m_x=-1;
  m_y=-1;
  m_z=-1;
  m_px=-1;
  m_py=-1;
  m_pz=-1;

 
  m_track_counts=0;
  SG::ReadHandle<TrackCollection> trackCollection {m_trackCollection, ctx};
  ATH_CHECK(trackCollection.isValid());
  for (const Trk::Track* track : *trackCollection)
  {
      //const Trk::TrackParameters* parameters = (*track->trackParameters())[0];
      // retrieve a pointer to the vector of parameters, dereference that point 
      //to get the vector, and then retrieve the first element of the vector

      if (track == nullptr) continue;
      std::set<std::pair<int, int>> layerMap;
      std::set<int> stationMap;

      // Check for hit in the three downstream stations
      for (auto measurement : *(track->measurementsOnTrack())) {
          const Tracker::FaserSCT_ClusterOnTrack* cluster = dynamic_cast<const Tracker::FaserSCT_ClusterOnTrack*>(measurement);
          if (cluster != nullptr) {
              Identifier id = cluster->identify();
              //int station = m_sctHelper->station(id);
              //int layer = m_sctHelper->layer(id);
              m_trackClusterIds.push_back(id);
          }
      }


      const Trk::TrackStateOnSurface* trackstate = (*track->trackStateOnSurfaces())[0];
      const Trk::TrackParameters* parameters = trackstate->trackParameters();

      pos = parameters->position();
      dir = parameters->momentum();
      dir.normalize();

      m_eta= parameters->eta();
      m_theta= 2*atan(exp(-1*m_eta));
      m_pt = parameters->pT();

      m_x=pos[0];
      m_y=pos[1];
      m_z=pos[2];
      m_px=dir[0];
      m_py=dir[1];
      m_pz=dir[2];

      m_chi2 = track->fitQuality()->chiSquared();

      m_trcluster_counts = track->fitQuality()->doubleNumberDoF();
      m_track_counts++;
  }

    //deion
  SG::ReadHandle<Tracker::FaserSCT_ClusterContainer> clusterContainer { m_clusterContainer, ctx };
  ATH_CHECK(clusterContainer.isValid());
  for (int i=0; i<5; i++) {m_layerspacepoints[i].clear();}
  SG::ReadHandle<FaserSCT_SpacePointContainer> spacePointContainer {m_spacePointContainerKey, ctx};
  ATH_CHECK(spacePointContainer.isValid());
  for (const FaserSCT_SpacePointCollection* spacePointCollection : *spacePointContainer) {
    m_nspacepoints += spacePointCollection->size();
    for (const Tracker::FaserSCT_SpacePoint *spacePoint: *spacePointCollection) {
      auto pos = spacePoint->globalPosition();
      //iterate over spacepoints
      //get cluster1, cluster2 for each spacepoint (or maybe just 1)
      //check if cluster is on "on track" list
      //if so, [4]=1

      double sp_z = pos.z();

      //#237-255,267-290,295-320
      int sp_layer=-1;
      if (sp_z>237 && sp_z<255){sp_layer=0;}
      else if (sp_z>267 && sp_z<290){sp_layer=1;}
      else if (sp_z>295 && sp_z<320){sp_layer=2;}

      m_layerspacepoints[0].push_back(sp_layer);
      m_layerspacepoints[1].push_back(pos.x());
      m_layerspacepoints[2].push_back(pos.y());
      m_layerspacepoints[3].push_back(sp_z);

      auto clusterID = spacePoint->cluster1()->identify();

      if(std::find(m_trackClusterIds.begin(), m_trackClusterIds.end(), clusterID) != m_trackClusterIds.end()) {
          m_layerspacepoints[4].push_back(1); //Cluster is on track
        } else {
          m_layerspacepoints[4].push_back(0); //Cluster is not on track
        }
    }
  }

    

  m_event_counts=1;

  m_tree->Fill();
  
  
  return StatusCode::SUCCESS;
}

StatusCode TrackerEfficiencyAlg::finalize() 
{
  return StatusCode::SUCCESS;
}

